open Term
open Overlap
open Rewriting
open Iroiro
open Arg
open Format

(* Global variables *)

let filenames  = ref []
let t          = ref None 
let tt         = ref "ttt2"
let smt        = ref "minismt"

(* Confleunce criteria *)

let exit_with_no () =
  printf "NO@."; exit 0

let kb rs = if Wcr.wcr_kb !tt rs then true else exit_with_no ()
let l rs =  if Rules.linear rs then true else exit_with_no ()
let ll rs = if Rules.left_linear rs then true else exit_with_no ()

let orthogonal rs =
  if Rules.left_linear rs && non_overlapping rs then true 
  else exit_with_no ()

let overlay rs = 
  List.for_all (fun (_,p,_,_) -> p = []) (overlap rs)

let weakly_orthogonal rs =
  if Rules.left_linear rs && 
     List.for_all (fun (s, t) -> s = t) (cp rs) then true
  else exit_with_no ()

let parallel_closed rs (s, t) = List.mem t (rewrite_parallel rs s)

let llpc rs =
  Rules.left_linear rs && 
  List.for_all (parallel_closed rs) (cp rs)

let cpssn rs = 
  if Rsn.not_sn (cps rs) rs then exit_with_no () else
  Sn.sn !tt (cps rs) rs 

let cps2sn rs = 
  let p (s, t) = s <> t in
  if Rsn.not_sn (cps ~p rs) rs then exit_with_no () else
  Sn.sn !tt (cps ~p rs) rs 

let dc n rs =
  Rules.left_linear rs &&
  not (Rsn.not_sn (cps rs) rs) &&
  Wcr.wcr_n n rs &&
  Sn.sn !tt (cps rs) rs 

let dc2 n rs =
  let p (s, t) = s <> t in
  Rules.left_linear rs &&
  not (Rsn.not_sn (cps ~p rs) rs) &&
  Wcr.wcr_n n rs && 
  Sn.sn !tt (cps ~p rs) rs

let dc2slow n rs =
  let p (s, t) = s <> t in
  Rules.left_linear rs &&
  Wcr.wcr_n n rs && 
  Sn.sn !tt (cps ~p rs) rs

let dc3 n rs =
  let p (s, t) = not (parallel_closed rs (s, t)) in
  Rules.left_linear rs &&
  not (Rsn.not_sn (cps ~p rs) rs) &&
  Wcr.wcr_n n rs && 
  Sn.sn !tt (cps ~p rs) rs

let (?:) r v = r := Some v
let (!:) r v = r := v
let prove = ref kb

let options = align [ 
  "-kb", 
     Unit (fun () -> prove := kb), 
     " Knuth-Bendix 1970";
  "-orthogonal",
     Unit (fun () -> prove := orthogonal), 
     " Orthogonal";
  "-overlay",
     Unit (fun () -> prove := overlay), 
     " Overlay";
  "-weakly_orthogonal",
     Unit (fun () -> prove := weakly_orthogonal), 
     " Weakly orthogonal";
  "-llpc",
     Unit (fun () -> prove := llpc), 
     " Left-linear parallel closed, Huet 80";
  "-cpssn",   
     Unit (fun n -> prove := cpssn), 
     " CPS(R)/R terminating or not";
  "-cps2sn",   
     Unit (fun n -> prove := cps2sn), 
     " CPS'(R)/R terminating or not";
  "-dc",   
     Int (fun n -> prove := dc n), 
     "<n> Decreasingly confluent with n-step joinability ";
  "-dc2",   
     Int (fun n -> prove := dc2 n), 
     "<n> Decreasingly confluent (wrt non-trivial CP) with n-step joinability";
  "-dc2slow",   
     Int (fun n -> prove := dc2slow n), 
     "<n> Decreasingly confluent (wrt non-trivial CP) with n-step joinability";
(*
  "-dc3",   
     Int (fun n -> prove := dc3 n), 
     "<n> Decreasingly confluent (wrt non-development-closed CP) with n-step joinability";
*)
  "-rlv",   
     Int (fun n -> prove := Rlv.solve ~tool:!smt n), 
     "<n> Rule labeling with n-step joinability";
  "-rl",   
     Int (fun n -> prove := Rl.solve ~tool:!smt n), 
     "<n> Rule labeling with n-step convertibility";
  "-t",    Float  ?:t, "<f> second timeout"; 
  "-tt",   String !:tt, "<tool> set termination tool (default: ttt2)"; 
  "-smt",  String !:smt, "<tool> set smt solver for QF_NIA (default: minismt)"; 

  "-linear", 
     Unit (fun () -> prove := l), 
     " check if TRS is linear";
  "-left_linear", 
     Unit (fun () -> prove := ll), 
     " check if TRS is left-linear";
  "-locally_confluent", 
     Int (fun n -> prove := Wcr.wcr_n n),
     "<n> check if TRS is WCR by n-step joinability";
] 
     
let short_usage = "Usage: saigawa <options> <file>\n"
let usage = "Confluence tool Saigawa v1.0\n" ^ short_usage ^ "Options are:"

let () = 
  parse options (fun it -> filenames := !filenames @ [it] ) usage;
  if List.length !filenames <> 1 then
    (eprintf "%s%!" short_usage; exit 1)
  else
    let filename = List.hd !filenames in
    let f_timeout () = printf "TIMEOUT@." in
    let f_prove () =
      let rs = Read.read_trs filename in
      if not (Rules.variable_condition rs) then
	(eprintf "Cannot handle TRS with extra variables.@."; exit 1);
      if !prove rs then 
	printf "YES@." 
      else 
	printf "MAYBE@." in
    Alarm.try_with_timeout ?timeout:!t f_timeout f_prove ()
